#include <algorithm>
#include "collision_detector.h"

using namespace std;


namespace ist {





void PointCollisionModel::transform(const matrix44 &trans)
{
	t_pos=trans*pos;

	bb.min.x=t_pos.x-radius;
	bb.max.x=t_pos.x+radius;

	bb.min.y=t_pos.y-radius;
	bb.max.y=t_pos.y+radius;

	bb.min.z=t_pos.z-radius;
	bb.max.z=t_pos.z+radius;
}


void LineCollisionModel::transform(const matrix44 &trans)
{
	t_head=trans*head;
	t_tail=trans*tail;

	bb.min.x = min<float>(t_head.x-radius, t_tail.x-radius);
	bb.max.x = max<float>(t_head.x+radius, t_tail.x+radius);

	bb.min.y = min<float>(t_head.y-radius, t_tail.y-radius);
	bb.max.y = max<float>(t_head.y+radius, t_tail.y+radius);

	bb.min.z = min<float>(t_head.z-radius, t_tail.z-radius);
	bb.max.z = max<float>(t_head.z+radius, t_tail.z+radius);

}


void PolygonCollisionModel::transform(const matrix44 &trans)
{
	if(chunc==0) return;

	center=trans*chunc->getBoundingSphere().pos;

	const std::vector<vector4> &vtx_array = chunc->getVertex();
	const std::vector<vector4> &nml_array = chunc->getNormal();

	for(unsigned int i=0; i<chunc->getVertexNum(); ++i)
		t_vertex[i]=trans*vtx_array[i];

	for(unsigned int i=0; i<chunc->getFaceNum(); ++i)
		t_normal[i]=trans*nml_array[i];

	bb=chunc->getBoundingBox().getAABB(trans);
}





CollisionDetector::CollisionDetector() : result(false), distance(0) {}

//|S-|S
template <>
bool CollisionDetector::detect_<PolygonCollisionModel, PolygonCollisionModel>
	(const PolygonCollisionModel& base, const PolygonCollisionModel& target)
{
	if(base.chunc==0 || target.chunc==0
	|| !base.bb.overlap(target.bb) )
		return false;

	const BasePolygonChunc::line_cont &line_array= base.chunc->getLines();
	for(BasePolygonChunc::line_cont::const_iterator p=line_array.begin(); p!=line_array.end(); ++p) {
		const vector4& head=base.t_vertex[ p->getIndex(0) ];
		const vector4& tail=base.t_vertex[ p->getIndex(1) ];

		for(unsigned int i=0; i<target.chunc->getFaceNum(); ++i) {
			const vector4& normal=target.t_normal[i];
			const vector4 *vtx[3] = {
				&target.t_vertex[ target.chunc->getIndex()[i*3+0] ],
				&target.t_vertex[ target.chunc->getIndex()[i*3+1] ],
				&target.t_vertex[ target.chunc->getIndex()[i*3+2] ],
			};
		
			float dist_tri = normal.dot( *vtx[0] );
			float dist_tail= normal.dot(tail);
			if( (normal.dot( head )-dist_tri)*(dist_tail-dist_tri)>0)
				continue;

			vector4 direction=(head-tail).normalize();
			float t=-(normal.dot(tail)-normal.dot(*vtx[0]))/normal.dot(direction);
			vector4 proj_pos=tail+direction*t;

			float inside=0;
			for(unsigned int j=0; j<3; ++j)
				inside+=(*vtx[j]-proj_pos).cos( *vtx[(j+1)%3]-proj_pos );

			if( inside>358.0f*radian ) {
				distance=base.chunc->getBoundingSphere().radius;
				collide_pos=proj_pos;
				collide_normal=normal;
				return true;
			}
		}
	}
	
	return false;
}

//|S-_
template <>
bool CollisionDetector::detect_<PolygonCollisionModel, PointCollisionModel>
	(const PolygonCollisionModel& base, const PointCollisionModel& target)
{
	if(base.chunc==0
	|| !base.bb.inner_point(target.t_pos, target.radius) )
		return false;

	const BasePolygonChunc::vertex_cont &vtx_array = base.t_vertex;
	const BasePolygonChunc::normal_cont &nml_array = base.t_normal;
	const BasePolygonChunc::index_cont  &idx_array = base.chunc->getIndex();
	const BasePolygonChunc::line_cont   &line_array= base.chunc->getLines();

	const LineIndex *nearest_line=0;
	float min_side_distance=0;


	for(BasePolygonChunc::line_cont::const_iterator p=line_array.begin();
		p!=line_array.end(); ++p) {	// ԋ߂̕ӂT

		const vector4 *vtx[2] = {
			&vtx_array[ p->getIndex(0) ],
			&vtx_array[ p->getIndex(1) ]
		};

		float u = ( target.t_pos-*vtx[0] ).dot( *vtx[1]-*vtx[0] ) / powf( (*vtx[1]-*vtx[0]).norm(), 2);

		if( u>0 && u<1 ) {
			float distance = ( target.t_pos-(*vtx[0]+(*vtx[1]-*vtx[0])*u) ).norm();

			if( distance<min_side_distance || nearest_line==0 ) {
				min_side_distance = distance;
				nearest_line = &(*p);
			}
		}
	}

	if(nearest_line==0)
		return false;


	bool is_inside=false;
	float min_distance=0;
	vector4 last_collide_pos;
	vector4 last_collide_normal;

	for(unsigned int i=0; i<2; ++i) {
		const int *triangle_index = nearest_line->getFace(i);
		if( triangle_index==0 )
			break;

		const vector4 *vtx[3] = {
			&vtx_array[ triangle_index[0] ],
			&vtx_array[ triangle_index[1] ],
			&vtx_array[ triangle_index[2] ],
		};

		const vector4& normal=nml_array[nearest_line->getFaceIndex(i)];


		for(int i=0; i<3; ++i) {
			float d=(*vtx[i]-target.t_pos).norm();
			if(d<target.radius) {
				distance=d;
				collide_pos = *vtx[i];
				collide_normal = normal;
				return true;
			}
		}

		float dist1=normal.dot( target.t_pos );	//_-_ԋ
		float dist2=normal.dot( *vtx[0] );		//_-ʊԋ
		float dist3=dist1-dist2;				//-_ԋAꂪʂ̔Αɂ

		if( fabsf(dist3)<fabsf(min_distance) || !is_inside ) {	//_Ɉԋ߂ʂT
			vector4 proj_pos = target.t_pos-normal*dist3;			//ʂɓeꂽ_̈ʒu

			float inside=0;
			for(unsigned int j=0; j<3; ++j)		//Op`Ƃ̓OBe_Ɛp360Ȃɂ͂B
				inside+=(*vtx[j]-proj_pos).cos( *vtx[(j+1)%3]-proj_pos );

			if(inside>358.0f*radian) {	//2̌덷F߂
				is_inside=true;
				min_distance=dist3;

				last_collide_pos=proj_pos;
				last_collide_normal=normal;
			}
		}
	}

	if(min_distance<target.radius && is_inside) {
		distance = target.radius;
		collide_pos = last_collide_pos;
		collide_normal = last_collide_normal;
		return true;
	}
	else
		return false;
}

//_-|S
template <>
bool CollisionDetector::detect_<PointCollisionModel, PolygonCollisionModel>
	(const PointCollisionModel& base, const PolygonCollisionModel& target)
{
	return detect_(target, base);
}

//|S-
template<>
bool CollisionDetector::detect_<PolygonCollisionModel, LineCollisionModel>
	(const PolygonCollisionModel& base, const LineCollisionModel& target)
{
	if( base.chunc==0
	|| !base.bb.overlap(target.bb) )
		return false;

	const BasePolygonChunc::vertex_cont &vtx_array = base.t_vertex;
	const BasePolygonChunc::normal_cont &nml_array = base.t_normal;
	const BasePolygonChunc::index_cont  &idx_array = base.chunc->getIndex();
	const BasePolygonChunc::line_cont   &line_array= base.chunc->getLines();
	const vector4& head=target.t_head;
	const vector4& tail=target.t_tail;

	bool is_inside=false;
	float min_distance=0;
	vector4 last_collide_pos;
	vector4 last_collide_normal;

	for(unsigned int i=0; i<base.chunc->getFaceNum(); ++i) {
		const vector4& normal=nml_array[i];
		const vector4 *vtx[3] = {
			&vtx_array[ idx_array[i*3+0] ],
			&vtx_array[ idx_array[i*3+1] ],
			&vtx_array[ idx_array[i*3+2] ],
		};
	
		float dist_tri = normal.dot( *vtx[0] );
		float dist_tail=normal.dot(tail);
		if( (normal.dot( head )-dist_tri)*(dist_tail-dist_tri)>0)
			continue;

		vector4 direction=(head-tail).normalize();
		float t=-(normal.dot(tail)-normal.dot(*vtx[0]))/normal.dot(direction);
		vector4 proj_pos=tail+direction*t;

		float inside=0;
		for(unsigned int j=0; j<3; ++j)
			inside+=(*vtx[j]-proj_pos).cos( *vtx[(j+1)%3]-proj_pos );

		if( (inside>358.0f*radian)
		&&  (fabsf(dist_tail)<fabsf(min_distance) || !is_inside) ) {
			is_inside=true;
			min_distance=dist_tail;
			last_collide_pos=proj_pos;
			last_collide_normal=normal;
		}
	}

	if(is_inside) {
		distance = target.radius+((target.t_tail-last_collide_pos).norm()+min_distance);
		collide_pos = last_collide_pos;
		collide_normal = last_collide_normal;
		return true;
	}
	else
		return false;
}

//-|S
template<>
bool CollisionDetector::detect_<LineCollisionModel, PolygonCollisionModel>
	(const LineCollisionModel& base, const PolygonCollisionModel& target)
{
	return detect_(target, base);
}



//_-_
template <>
bool CollisionDetector::detect_<PointCollisionModel, PointCollisionModel>
	(const PointCollisionModel& base, const PointCollisionModel& target)
{
	if( !base.bb.overlap(target.bb) )
		return false;

	vector4 pos_gap=target.t_pos-base.t_pos;
	float union_radius=base.radius + target.radius;

	if(pos_gap.norm() < union_radius) {
		distance=union_radius;
		collide_normal = vector4(pos_gap).normalize();
		collide_pos = base.t_pos;
		return true;
	}

	return false;
}


//-_
template <>
bool CollisionDetector::detect_<LineCollisionModel, PointCollisionModel>
	(const LineCollisionModel& base, const PointCollisionModel& target)
{
	if( !base.bb.overlap(target.bb) )
		return false;

	bool _result=false;
	const vector4 &head=base.t_head;
	const vector4 &tail=base.t_tail;
	vector4 segment=head-tail;

	float union_radius=base.radius+target.radius;
	float u=0;
	if(tail!=head)
		u=( target.t_pos-tail ).dot( segment ) / powf( segment.norm(), 2);

	if(u>0 && u<1) {
		vector4 nearest=tail+(segment*u);
		float distance = (target.t_pos-nearest).norm();

		if( distance < union_radius ) {
			collide_normal = (target.t_pos-nearest).normalize();
			collide_pos = nearest;
			_result=true;
		}
	}
	else if((target.t_pos-head).norm()<union_radius) {
		collide_normal = (target.t_pos-head).normalize();
		collide_pos = head;
		_result=true;
	}
	else if((target.t_pos-tail).norm()<union_radius) {
		collide_normal = (target.t_pos-tail).normalize();
		collide_pos = tail;
		_result=true;
	}

	if(_result)
		distance=union_radius;

	return _result;
}


//_-
template <>
bool CollisionDetector::detect_<PointCollisionModel, LineCollisionModel>
	(const PointCollisionModel& base, const LineCollisionModel& target)
{
	return detect_(target, base);
}


//-
template <>
bool CollisionDetector::detect_<LineCollisionModel, LineCollisionModel>
	(const LineCollisionModel& base, const LineCollisionModel& target)
{
	if( !base.bb.overlap(target.bb) )
		return false;

	return false;
}


//NX^
bool CollisionDetector::detect(const BaseCollisionModel &base, const BaseCollisionModel &target)
{
	using namespace collision_type;

	if( base.type()==POINT ) {
		if     ( target.type()==POINT )
			result=detect_( (static_cast<const PointCollisionModel&>(base)),
							(static_cast<const PointCollisionModel&>(target)) );
		else if( target.type()==LINE )
			result=detect_( (static_cast<const PointCollisionModel&>(base)),
							(static_cast<const LineCollisionModel&>(target)) );
		else if( target.type()==POLYGON )
			result=detect_( (static_cast<const PointCollisionModel&>(base)),
							(static_cast<const PolygonCollisionModel&>(target)) );
		else
			result=false;
	}
	else if( base.type()==LINE ) {
		if     ( target.type()==POINT )
			result=detect_( (static_cast<const LineCollisionModel&>(base)),
							(static_cast<const PointCollisionModel&>(target)) );
		else if( target.type()==LINE )
			result=detect_( (static_cast<const LineCollisionModel&>(base)),
							(static_cast<const LineCollisionModel&>(target)) );
		else if( target.type()==POLYGON )
			result=detect_( (static_cast<const LineCollisionModel&>(base)),
							(static_cast<const PolygonCollisionModel&>(target)) );
		else
			result=false;
	}
	else if( base.type()==POLYGON ) {
		if( target.type()==POINT )
			result=detect_( (static_cast<const PolygonCollisionModel&>(base)),
							(static_cast<const PointCollisionModel&>(target)) );
		else if( target.type()==LINE )
			result=detect_( (static_cast<const PolygonCollisionModel&>(base)),
							(static_cast<const LineCollisionModel&>(target)) );
		else if( target.type()==POLYGON )
			result=detect_( (static_cast<const PolygonCollisionModel&>(base)),
							(static_cast<const PolygonCollisionModel&>(target)) );
		else
			result=false;
	}
	else {
		result=false;
	}

	return result;
}



} // ist

