#include <algorithm>
#include "spline.h"
using namespace std;



namespace ist {

template<typename T>
struct less_equal_elem_x {
	bool operator()(const T& lhs, const T& rhs) {
		return lhs.pos.x <= rhs.pos.x;
	}
};


void Spline2D::insert(const vector2& p)
{
	points.push_back( point_t(p) );

	range_t r=get_edge(p.x);
	if(r.first!=end() && r.first->out.x > p.x)
		r.first->out.x = p.x;
	if(r.second!=end() && r.second->in.x < p.x)
		r.second->in.x = p.x;

	std::sort(points.begin(), points.end(), less_equal_elem_x<Spline2D::point_t>());
}

void Spline2D::insert(const vector2& i, const vector2& p, const vector2& o)
{
	points.push_back( point_t(i, p, o) );
	std::sort(points.begin(), points.end(), less_equal_elem_x<Spline2D::point_t>());
}

Spline2D::range_t Spline2D::get_edge(float x)
{
	range_t pi;

	if(points.empty()) {
		pi.first=end();
		pi.second=end();
	}
	else if(points.size()==1) {
		if(x > points.front().pos.x) {
			pi.first=end();
			pi.second=begin();
		}
		else {
			pi.first=begin();
			pi.second=end();
		}
	}
	else {
		pi=equal_range(begin(), end(), point_t(vector2(x, 0)), less_equal_elem_x<Spline2D::point_t>());
		--pi.first;
	}

	return pi;
}

vector2 Spline2D::linear(float x)
{
	if(points.empty())
		return vector2();
	else if(x <= points.front().pos.x)
		return vector2(x, points.front().pos.y);
	else if(x >= points.back().pos.x)
		return vector2(x, points.back().pos.y);

	range_t range=get_edge(x);
	point_t& prev=*range.first;
	point_t& next=*range.second;

	float dir=(next.pos.y - prev.pos.y)/(next.pos.x - prev.pos.x);
	return vector2(x, prev.pos.y + dir*(x-prev.pos.x));
}

vector2 Spline2D::bezier(float x)
{
	if(points.empty())
		return vector2();
	else if(x <= points.front().pos.x)
		return vector2(x, points.front().pos.y);
	else if(x >= points.back().pos.x)
		return vector2(x, points.back().pos.y);

	range_t range=get_edge(x);
	point_t& prev=*range.first;
	point_t& next=*range.second;

	int sw=0;
	if(next.pos==next.in)  sw|=1;
	if(prev.pos==prev.out) sw|=2;

	float u = (x - prev.pos.x) / (next.pos.x - prev.pos.x);

	if(sw!=0) {
		if(sw==3) {
			float dir=(next.pos.y-prev.pos.y)/(next.pos.x-prev.pos.x);
			return vector2(x, prev.pos.y + dir*(x-prev.pos.x));
		}
		else {
			float w[3] = {
			(1-u)*(1-u),
			    u*(1-u) * 2,
			    u*   u,
			};


			if(sw==1)
				return vector2(
					(w[0]*prev.pos.x + w[1]*prev.out.x + w[2]*next.pos.x),
					(w[0]*prev.pos.y + w[1]*prev.out.y + w[2]*next.pos.y));
			if(sw==2)
				return vector2(
					(w[0]*prev.pos.x + w[1]*next.in.x + w[2]*next.pos.x),
					(w[0]*prev.pos.y + w[1]*next.in.y + w[2]*next.pos.y));
		}
	}
	else {
		float w[4] = {
		(1-u)*(1-u)*(1-u),
		    u*(1-u)*(1-u)*3,
		    u*    u*(1-u)*3,
		    u*    u*   u,
		};


		return vector2(
			(w[0]*prev.pos.x + w[1]*prev.out.x + w[2]*next.in.x + w[3]*next.pos.x),
			(w[0]*prev.pos.y + w[1]*prev.out.y + w[2]*next.in.y + w[3]*next.pos.y));
	}

	return vector2();
}


}

