#include <boost/shared_ptr.hpp>
#include <map>
#include <SDL.h>
#include "ist_gl.h"
#include "ist_lua.h"
#include "static_font.h"
#include "timeline.h"

using namespace std;
using namespace boost;
using namespace ist;

typedef shared_ptr<Timeline1> timeline1_ptr;
typedef shared_ptr<Timeline3> timeline3_ptr;
typedef map<string, timeline1_ptr> timeline1_map;
typedef map<string, timeline3_ptr> timeline3_map;


static timeline1_map tl1_map;
static timeline3_map tl3_map;


namespace timeline {
void add_timeline1(const string& name, Timeline1 *tl)
{
	tl1_map.insert( pair<string, timeline1_ptr>(name, timeline1_ptr(tl)) );
}

void add_timeline3(const string& name, Timeline3 *tl)
{
	tl3_map.insert( pair<string, timeline3_ptr>(name, timeline3_ptr(tl)) );
}
} // namespace timeline




class MainFrame
{
public:
	MainFrame();
	MainFrame(const char *fn);
	~MainFrame();
	int mainLoop();
	void draw();
	
	void eventResize(SDL_ResizeEvent& resize);
	void eventMouse(SDL_MouseButtonEvent& button);
	void eventMouse(SDL_MouseMotionEvent& motion);
	void eventKeyboard(SDL_KeyboardEvent& key);

	void drawGuideMessage();
	void drawSystemMessage();
	void reload();
	bool load(const char *fn);
	void save();

	Timeline1* getNearLine(float r=8);
	pair<Timeline1*, Timeline1::iterator> getNearPoint(float r=8);

private:
	enum {
		EM_NORMAL=0,
		EM_INSERT=1,
		EM_DELETE=2,
	};

	void init();

	int window_width;
	int window_height;
	float screen_region[4];
	int edit_mode;
	int mouse_state;
	bool toggle_help;
	bool toggle_shift;
	bool toggle_ctrl;
	vector2 mouse_pos;
	Timeline1::iterator focus_point;
	Timeline1 *focus_tl1;
	string current_tl3_name;
	string current_tl1_name;
	string filename;
	StaticFont font;
};


MainFrame::MainFrame()
{
	init();

	char t_filemane[128];
	int i=0;
	while(1) {
		sprintf(t_filemane, "tl3_%03d.lua", i);
		FILE *check=fopen(t_filemane, "r");
		if (check==NULL) {
			break;
		}
		else {
			fclose(check);
			i++;
		}
	}

	filename=t_filemane;
	current_tl3_name="no_name";
	tl3_map[current_tl3_name]=timeline3_ptr(new Timeline3);
	tl3_map[current_tl3_name]->x.insert(vector2(0, 10));
	tl3_map[current_tl3_name]->x.insert(vector2(100, 10));
	tl3_map[current_tl3_name]->y.insert(vector2(0, 0));
	tl3_map[current_tl3_name]->y.insert(vector2(100, 0));
	tl3_map[current_tl3_name]->z.insert(vector2(0, -10));
	tl3_map[current_tl3_name]->z.insert(vector2(100, -10));
}

MainFrame::MainFrame(const char *fn)
{
	init();
	load(fn);
}

MainFrame::~MainFrame()
{
	SDL_Quit();
}

void MainFrame::init()
{
	window_width=800;
	window_height=500;
	screen_region[0]=0;
	screen_region[1]=window_width;
	screen_region[2]=-window_height/2;
	screen_region[3]=window_height/2;

	mouse_state=0;

	edit_mode=EM_NORMAL;
	toggle_help=true;
	toggle_shift=false;
	toggle_ctrl=false;

	focus_point=Timeline1::iterator();
	focus_tl1=0;

	if( SDL_Init( SDL_INIT_VIDEO )<0 )
		throw Error(SDL_GetError());

	const SDL_VideoInfo *info=SDL_GetVideoInfo();
	int bpp=info->vfmt->BitsPerPixel;

	int flag=SDL_OPENGL | SDL_RESIZABLE;
	SDL_GL_SetAttribute( SDL_GL_RED_SIZE, 5 );
	SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, 5 );
	SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, 5 );
	SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, 16 );
	SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );

	SDL_SetVideoMode( window_width, window_height, bpp, flag );
}

bool MainFrame::load(const char *fn)
{
	filename=fn;
	reload();
	return true;
}



void MainFrame::reload()
{
	tl1_map.clear();
	tl3_map.clear();

	focus_point=Timeline1::iterator();
	focus_tl1=0;

	if(!filename.empty()) {
		lua::Thread lthread;
		timeline::regist_timeline_functions(lthread);
		lua_dofile(lthread, filename.c_str());
	}
	if(!tl3_map.empty()) {
		current_tl3_name = tl3_map.begin()->first;
	}
}


static char* create_point_string(Timeline1::point_t& p)
{
	static char buf[512];
	sprintf(buf, "vector2.new(%.2f, %.2f), vector2.new(%.2f, %.2f), vector2.new(%.2f, %.2f)",
		p.in.x, p.in.y,  p.pos.x, p.pos.y,  p.out.x, p.out.y);

	return buf;
}

void MainFrame::save()
{
	FILE *out=0;
	if(filename.empty())
		return;
	else
		out=fopen(filename.c_str(), "w");

	for(timeline3_map::iterator p=tl3_map.begin(); p!=tl3_map.end(); ++p) {
		fprintf(out, "timeline3_begin(\"%s\")\n", p->first.c_str());

		for(Timeline1::iterator q=p->second->x.begin(); q!=p->second->x.end(); ++q)
			fprintf(out, "timeline3_insert_x(%s)\n", create_point_string(*q));
		for(Timeline1::iterator q=p->second->y.begin(); q!=p->second->y.end(); ++q)
			fprintf(out, "timeline3_insert_y(%s)\n", create_point_string(*q));
		for(Timeline1::iterator q=p->second->z.begin(); q!=p->second->z.end(); ++q)
			fprintf(out, "timeline3_insert_z(%s)\n", create_point_string(*q));

		fprintf(out, "timeline3_end()\n\n");
	}
	fclose(out);
}



int MainFrame::mainLoop() {
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

	glClearColor( 0, 0, 0, 0 );
	glViewport( 0, 0, window_width, window_height );

	draw();

	SDL_Event event;
	while( SDL_WaitEvent(&event) ) {
		int result=0;

		switch( event.type ) {
		case SDL_QUIT:
			result=1;
			break;
		case SDL_VIDEORESIZE:
			eventResize(event.resize);
			result=2;
			break;
		case SDL_KEYDOWN:
			eventKeyboard(event.key);
			result=2;
			if(event.key.type==SDL_KEYDOWN && event.key.keysym.sym==SDLK_ESCAPE)
				result=1;
			break;
		case SDL_MOUSEBUTTONDOWN:
			eventMouse(event.button);
			result=2;
			break;
		case SDL_MOUSEBUTTONUP:
			eventMouse(event.button);
			result=2;
			break;
		case SDL_MOUSEMOTION:
			eventMouse(event.motion);
			result=2;
			break;
		default:
			break;
		}

		if(result==1)
			break;
		else if(result==2)
			draw();
	}

	return 0;
}


void MainFrame::eventResize(SDL_ResizeEvent& resize)
{
	window_width=resize.w;
	window_height=resize.h;
	screen_region[0]=0;
	screen_region[1]=window_width;
	screen_region[2]=-window_height/2;
	screen_region[3]=window_height/2;

	int flag=SDL_OPENGL | SDL_RESIZABLE;
	SDL_SetVideoMode( window_width, window_height, 0, flag );
	glViewport( 0, 0, window_width, window_height );
}

Timeline1* MainFrame::getNearLine(float r)
{
	if(current_tl3_name.empty())
		return 0;

	Timeline3& tl3=*tl3_map[current_tl3_name];

	if( (vector2(mouse_pos.x, tl3.x.getValue(mouse_pos.x)) - mouse_pos).square() < r*r ) {
		return &tl3.x;
	}
	if( (vector2(mouse_pos.x, tl3.y.getValue(mouse_pos.x)) - mouse_pos).square() < r*r ) {
		return &tl3.y;
	}
	if( (vector2(mouse_pos.x, tl3.z.getValue(mouse_pos.x)) - mouse_pos).square() < r*r ) {
		return &tl3.z;
	}

	return 0;
}

pair<Timeline1*, Timeline1::iterator> MainFrame::getNearPoint(float r)
{
	if(current_tl3_name.empty())
		return pair<Timeline1*, Timeline1::iterator>(0, Timeline1::iterator());

	Timeline3& tl3=*tl3_map[current_tl3_name];

	for(Timeline1::iterator p=tl3.x.begin(); p!=tl3.x.end(); ++p) {
		if( (p->pos - mouse_pos).square() < r*r ) {
			return pair<Timeline1*, Timeline1::iterator>(&tl3.x, p);
		}
	}
	for(Timeline1::iterator p=tl3.y.begin(); p!=tl3.y.end(); ++p) {
		if( (p->pos - mouse_pos).square() < r*r ) {
			return pair<Timeline1*, Timeline1::iterator>(&tl3.y, p);
		}
	}
	for(Timeline1::iterator p=tl3.z.begin(); p!=tl3.z.end(); ++p) {
		if( (p->pos - mouse_pos).square() < r*r ) {
			return pair<Timeline1*, Timeline1::iterator>(&tl3.z, p);
		}
	}

	return pair<Timeline1*, Timeline1::iterator>(0, Timeline1::iterator());
}

void MainFrame::eventMouse(SDL_MouseButtonEvent& button)
{
	mouse_state=SDL_GetMouseState(0, 0);

	if(current_tl3_name.empty())
		return;


	if(edit_mode==EM_NORMAL && button.type==SDL_MOUSEBUTTONDOWN ) {
		pair<Timeline1*, Timeline1::iterator> result=getNearPoint();

		if(button.button==SDL_BUTTON_MIDDLE) {
			focus_tl1=result.first;
			focus_point=result.second;
		}
		else if(button.button==SDL_BUTTON_LEFT) {
			if(result.first!=0 && result.second!=Timeline1::iterator()) {
				focus_tl1=result.first;
				focus_point=result.second;
			}
		}
	}
	else if(edit_mode==EM_NORMAL && button.type==SDL_MOUSEBUTTONUP ) {
		if(focus_tl1!=0 && focus_point!=Timeline1::iterator()) {
			if(button.button==SDL_BUTTON_LEFT && ((focus_point->in - focus_point->pos).square() < 8*8)) {
				 focus_point->in = focus_point->pos;
			}
			else if(button.button==SDL_BUTTON_RIGHT && ((focus_point->out - focus_point->pos).square() < 8*8)) {
				 focus_point->out = focus_point->pos;
			}
		}
	}


	if(edit_mode==EM_DELETE && button.type==SDL_MOUSEBUTTONDOWN) {
		pair<Timeline1*, Timeline1::iterator> result=getNearPoint();
		if(result.first!=0 && result.second!=Timeline1::iterator()) {
			result.first->erase(result.second);
		}
	}

	if(edit_mode==EM_INSERT && button.type==SDL_MOUSEBUTTONDOWN) {
		Timeline1 *near_line=getNearLine();
		if(near_line!=0) {
			focus_point=Timeline1::iterator();
			near_line->insert(vector2(mouse_pos.x, near_line->getValue(mouse_pos.x)));
		}
	}
}

void MainFrame::eventMouse(SDL_MouseMotionEvent& motion)
{
	float x_rate = static_cast<float>(motion.x)/static_cast<float>(window_width);
	float y_rate = static_cast<float>(motion.y)/static_cast<float>(window_height);
	mouse_pos.x = screen_region[0]+(screen_region[1]-screen_region[0])*x_rate;
	mouse_pos.y = screen_region[3]-(screen_region[3]-screen_region[2])*y_rate;

	if(current_tl3_name.empty())
		return;

	if(focus_point!=Timeline1::iterator()) {
		if(mouse_state & 2) {
			focus_point->pos.x+=motion.xrel; focus_point->pos.y-=motion.yrel;
			focus_point->in.x+=motion.xrel;  focus_point->in.y-=motion.yrel;
			focus_point->out.x+=motion.xrel; focus_point->out.y-=motion.yrel;
		}
		else if(mouse_state & 1) { focus_point->in.x+=motion.xrel;  focus_point->in.y-=motion.yrel; }
		else if(mouse_state & 4) { focus_point->out.x+=motion.xrel; focus_point->out.y-=motion.yrel; }
		tl3_map[current_tl3_name]->sampling();
	}
	else if(mouse_state & 2) {
		float move_x=((screen_region[1]-screen_region[0])/300.0f)*motion.xrel;
		float move_y=((screen_region[3]-screen_region[4])/300.0f)*motion.yrel;
		screen_region[0]+=move_x;
		screen_region[1]+=move_x;
		screen_region[2]-=move_y;
		screen_region[3]-=move_y;
	}
}


void MainFrame::eventKeyboard(SDL_KeyboardEvent &key)
{
	float scroll_speed=50;
	float scale_rate=0.1;
	vector2 center((screen_region[1]-screen_region[0])/2, (screen_region[3]-screen_region[2])/2);

	if(key.type==SDL_KEYDOWN) {
		switch(key.keysym.sym) {
		case SDLK_h:
			toggle_help=!toggle_help;
			break;
		case SDLK_r:
			reload();
			break;
		case SDLK_s:
			save();
			break;
		case SDLK_e:
			edit_mode=EM_NORMAL;
			break;
		case SDLK_i:
			edit_mode=EM_INSERT;
			break;
		case SDLK_d:
			edit_mode=EM_DELETE;
			break;
		case SDLK_DELETE:
			if(focus_tl1!=0 && focus_point!=Timeline1::iterator()) {
				focus_tl1->erase(focus_point);
				focus_point=Timeline1::iterator();
			}
			break;

		case SDLK_LEFT:
			screen_region[0]-=scroll_speed;
			screen_region[1]-=scroll_speed;
			break;
		case SDLK_RIGHT:
			screen_region[0]+=scroll_speed;
			screen_region[1]+=scroll_speed;
			break;
		case SDLK_UP:
			screen_region[2]+=scroll_speed;
			screen_region[3]+=scroll_speed;
			break;
		case SDLK_DOWN:
			screen_region[2]-=scroll_speed;
			screen_region[3]-=scroll_speed;
			break;
		case SDLK_SEMICOLON:
			if(toggle_shift) {
				screen_region[0]=screen_region[0]+((center.x-screen_region[0])*scale_rate);
				screen_region[1]=screen_region[1]+((center.x-screen_region[1])*scale_rate);
				screen_region[2]=screen_region[2]+((center.y-screen_region[2])*scale_rate);
				screen_region[3]=screen_region[3]+((center.y-screen_region[3])*scale_rate);
			}
			break;
		case SDLK_MINUS:
			screen_region[0]=screen_region[0]-((center.x-screen_region[0])*scale_rate);
			screen_region[1]=screen_region[1]-((center.x-screen_region[1])*scale_rate);
			screen_region[2]=screen_region[2]-((center.y-screen_region[2])*scale_rate);
			screen_region[3]=screen_region[3]-((center.y-screen_region[3])*scale_rate);
			break;
		case SDLK_RSHIFT:
		case SDLK_LSHIFT:
			toggle_shift=true;
			break;
		case SDLK_RCTRL:
		case SDLK_LCTRL:
			toggle_ctrl=true;
			break;
		default:
			break;
		}
	}

	if(key.type==SDL_KEYUP) {
		switch(key.keysym.sym) {
		case SDLK_RSHIFT:
		case SDLK_LSHIFT:
			toggle_shift=false;
			break;
		case SDLK_RCTRL:
		case SDLK_LCTRL:
			toggle_ctrl=false;
			break;
		default:
			break;
		}
	}
}



template<class Iter>
void draw_points(Iter begin, Iter end)
{
	glBegin(GL_POINTS);
	for(; begin!=end; ++begin) {
		glVertex2fv( begin->pos.v );
	}
	glEnd();
}

template<class Iter>
void draw_lines(Iter begin, Iter end)
{
	glBegin(GL_LINE_STRIP);
	for(; begin!=end; ++begin) {
		glVertex2fv( begin->v );
	}
	glEnd();
}

void draw_detail(Timeline1::point_t& p)
{
	glBegin(GL_POINTS);
	glVertex2fv( p.pos.v );
	glVertex2fv( p.in.v );
	glVertex2fv( p.out.v );
	glEnd();
	glBegin(GL_LINE_STRIP);
	glVertex2fv( p.in.v );
	glVertex2fv( p.pos.v );
	glVertex2fv( p.out.v );
	glEnd();
}

void MainFrame::draw()
{
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(screen_region[0], screen_region[1], screen_region[2], screen_region[3], 1, 50000);
	glMatrixMode( GL_MODELVIEW );

	gl::Camera camera;
	camera.pos=vector4(0, 0, 100);
	camera.target=vector4(0, 0, 0);
	camera.look();

	if(!current_tl3_name.empty()) {
		Timeline3& tl3=*tl3_map[current_tl3_name];

		glColor4ub(255, 0, 0, 255);
		draw_lines( tl3.x.s_points.begin(), tl3.x.s_points.end() );

		glColor4ub(0, 255, 0, 255);
		draw_lines( tl3.y.s_points.begin(), tl3.y.s_points.end() );

		glColor4ub(0, 0, 255, 255);
		draw_lines( tl3.z.s_points.begin(), tl3.z.s_points.end() );

		glPointSize(7);
		glColor4ub(128, 128, 128, 255);
		draw_points( tl3.x.begin(), tl3.x.end() );
		draw_points( tl3.y.begin(), tl3.y.end() );
		draw_points( tl3.z.begin(), tl3.z.end() );

		if(focus_point!=Timeline1::iterator()) {
			glPointSize(10);
			glColor4ub(255, 255, 255, 255);
			glBegin(GL_POINTS);
			draw_detail( *focus_point );
		}
	}

	drawGuideMessage();
	drawSystemMessage();

	SDL_GL_SwapBuffers();
}


static void draw_font( gl::Kanji& kanji, int x, int y, const char *str)
{
	gl::Color back(0, 0, 0, 1);
	back.assign();
	kanji.print(x+1, y, str);
	kanji.print(x-1, y, str);
	kanji.print(x, y+1, str);
	kanji.print(x, y-1, str);

	gl::Color front(1, 1, 1, 1);
	front.assign();
	kanji.print(x, y, str);
}

void MainFrame::drawSystemMessage()
{
	char str[256];

	if(focus_point!=Timeline1::iterator()) {
		sprintf(str, "%.2f, %.2f", focus_point->pos.x, focus_point->pos.y);
		draw_font(font, 5, window_height-40, str);
	}

	switch(edit_mode) {
	case EM_NORMAL:
		sprintf(str, "--EDIT--");
		break;
	case EM_INSERT:
		sprintf(str, "--INSERT--");
		break;
	case EM_DELETE:
		sprintf(str, "--DELETE--");
		break;
	}
	draw_font(font, 5, window_height-20, str);
}

void MainFrame::drawGuideMessage()
{
	if(!toggle_help)
		return;

	static const string message1[]={
		"~hhbO: |Cg̈ړAʂ̃XN[",
		"E/hbO: ڐ̈ړ",
		"h: ̃bZ[W̕\/\",
		"s: ۑ",
		"r: ēǂݍ",
		"e: GfBbg[h",
		"i: CT[g[h",
		"d: f[g[h",
		"+/-: g/k",
	};

	for(int i=0; i<(sizeof(message1)/sizeof(string)); ++i) {
		draw_font(font, 5, 5+20*i, message1[i].c_str());
	}
}



int main(int argc, char *argv[])
{
	try {
		MainFrame mframe;
		if(argc==2)
			mframe.load(argv[1]);

		return mframe.mainLoop();
	}
	catch(Error& err) {
		printf("%s", err.what());
		return 1;
	}
}
